Class: PerfectShape::CubicBezierCurve
- Includes:
- MultiPoint
- Defined in:
- lib/perfect_shape/cubic_bezier_curve.rb
Overview
Mostly ported from java.awt.geom: https://docs.oracle.com/javase/8/docs/api/java/awt/geom/QuadCurve2D.html
Constant Summary collapse
- OUTLINE_MINIMUM_DISTANCE_THRESHOLD =
BigDecimal('0.001')
Instance Attribute Summary
Attributes included from MultiPoint
Class Method Summary collapse
-
.point_crossings(x1, y1, xc1, yc1, xc2, yc2, x2, y2, px, py, level = 0) ⇒ Object
Calculates the number of times the cubic bézier curve from (x1,y1) to (x2,y2) crosses the ray extending to the right from (x,y).
Instance Method Summary collapse
-
#contain?(x_or_point, y = nil, outline: false, distance_tolerance: 0) ⇒ @code true
Checks if cubic bézier curve contains point (two-number Array or x, y args).
-
#curve_center_point ⇒ Object
The center point on the outline of the curve.
-
#curve_center_x ⇒ Object
The center point x on the outline of the curve.
-
#curve_center_y ⇒ Object
The center point y on the outline of the curve.
-
#point_crossings(x_or_point, y = nil, level = 0) ⇒ Object
Calculates the number of times the cubic bézier curve crosses the ray extending to the right from (x,y).
- #point_segment_distance(x_or_point, y = nil, minimum_distance_threshold: OUTLINE_MINIMUM_DISTANCE_THRESHOLD) ⇒ Object
-
#subdivisions(number = 2) ⇒ Object
Subdivides CubicBezierCurve exactly at its curve center returning 2 CubicBezierCurve's as a two-element Array by default
numberparameter may be specified as an even number in case more subdivisions are needed.
Methods included from MultiPoint
#initialize, #max_x, #max_y, #min_x, #min_y
Methods inherited from Shape
#==, #bounding_box, #center_x, #center_y, #height, #max_x, #max_y, #min_x, #min_y, #normalize_point, #width
Class Method Details
.point_crossings(x1, y1, xc1, yc1, xc2, yc2, x2, y2, px, py, level = 0) ⇒ Object
Calculates the number of times the cubic bézier curve from (x1,y1) to (x2,y2) crosses the ray extending to the right from (x,y). If the point lies on a part of the curve, then no crossings are counted for that intersection. the level parameter should be 0 at the top-level call and will count up for each recursion level to prevent infinite recursion +1 is added for each crossing where the Y coordinate is increasing -1 is added for each crossing where the Y coordinate is decreasing
37 38 39 40 41 42 43 44 45 46 47 48 49 50 51 52 53 54 55 56 57 58 59 60 61 62 63 64 65 66 67 68 69 70 71 72 |
# File 'lib/perfect_shape/cubic_bezier_curve.rb', line 37 def point_crossings(x1, y1, xc1, yc1, xc2, yc2, x2, y2, px, py, level = 0) return 0 if (py < y1 && py < yc1 && py < yc2 && py < y2) return 0 if (py >= y1 && py >= yc1 && py >= yc2 && py >= y2) # Note y1 could equal yc1... return 0 if (px >= x1 && px >= xc1 && px >= xc2 && px >= x2) if (px < x1 && px < xc1 && px < xc2 && px < x2) if (py >= y1) return 1 if (py < y2) else # py < y1 return -1 if (py >= y2) end # py outside of y12 range, and/or y1==yc1 return 0 end # double precision only has 52 bits of mantissa return PerfectShape::Line.point_crossings(x1, y1, x2, y2, px, py) if (level > 52) xmid = BigDecimal((xc1 + xc2).to_s) / 2; ymid = BigDecimal((yc1 + yc2).to_s) / 2; xc1 = BigDecimal((x1 + xc1).to_s) / 2; yc1 = BigDecimal((y1 + yc1).to_s) / 2; xc2 = BigDecimal((xc2 + x2).to_s) / 2; yc2 = BigDecimal((yc2 + y2).to_s) / 2; xc1m = BigDecimal((xc1 + xmid).to_s) / 2; yc1m = BigDecimal((yc1 + ymid).to_s) / 2; xmc1 = BigDecimal((xmid + xc2).to_s) / 2; ymc1 = BigDecimal((ymid + yc2).to_s) / 2; xmid = BigDecimal((xc1m + xmc1).to_s) / 2; ymid = BigDecimal((yc1m + ymc1).to_s) / 2; # [xy]mid are NaN if any of [xy]c0m or [xy]mc1 are NaN # [xy]c0m or [xy]mc1 are NaN if any of [xy][c][01] are NaN # These values are also NaN if opposing infinities are added return 0 if (xmid.nan? || ymid.nan?) point_crossings(x1, y1, xc1, yc1, xc1m, yc1m, xmid, ymid, px, py, level+1) + point_crossings(xmid, ymid, xmc1, ymc1, xc2, yc2, x2, y2, px, py, level+1) end |
Instance Method Details
#contain?(x_or_point, y = nil, outline: false, distance_tolerance: 0) ⇒ @code true
Checks if cubic bézier curve contains point (two-number Array or x, y args)
the cubic bézier curve, false if the point lies outside of the cubic bézier curve's bounds.
88 89 90 91 92 93 94 95 96 97 98 99 100 101 102 103 104 105 106 107 108 109 110 111 |
# File 'lib/perfect_shape/cubic_bezier_curve.rb', line 88 def contain?(x_or_point, y = nil, outline: false, distance_tolerance: 0) x, y = normalize_point(x_or_point, y) return unless x && y if outline minimum_distance_threshold = OUTLINE_MINIMUM_DISTANCE_THRESHOLD + distance_tolerance point_segment_distance(x, y, minimum_distance_threshold: minimum_distance_threshold) < minimum_distance_threshold else # Either x or y was infinite or NaN. # A NaN always produces a negative response to any test # and Infinity values cannot be "inside" any path so # they should return false as well. return false if (!(x * 0.0 + y * 0.0 == 0.0)) # We count the "Y" crossings to determine if the point is # inside the curve bounded by its closing line. x1 = points[0][0] y1 = points[0][1] x2 = points[3][0] y2 = points[3][1] line = PerfectShape::Line.new(points: [[x1, y1], [x2, y2]]) crossings = line.point_crossings(x, y) + point_crossings(x, y); (crossings & 1) == 1 end end |
#curve_center_point ⇒ Object
The center point on the outline of the curve
128 129 130 |
# File 'lib/perfect_shape/cubic_bezier_curve.rb', line 128 def curve_center_point subdivisions.last.points[0] end |
#curve_center_x ⇒ Object
The center point x on the outline of the curve
133 134 135 |
# File 'lib/perfect_shape/cubic_bezier_curve.rb', line 133 def curve_center_x subdivisions.last.points[0][0] end |
#curve_center_y ⇒ Object
The center point y on the outline of the curve
138 139 140 |
# File 'lib/perfect_shape/cubic_bezier_curve.rb', line 138 def curve_center_y subdivisions.last.points[0][1] end |
#point_crossings(x_or_point, y = nil, level = 0) ⇒ Object
Calculates the number of times the cubic bézier curve crosses the ray extending to the right from (x,y). If the point lies on a part of the curve, then no crossings are counted for that intersection. the level parameter should be 0 at the top-level call and will count up for each recursion level to prevent infinite recursion +1 is added for each crossing where the Y coordinate is increasing -1 is added for each crossing where the Y coordinate is decreasing
121 122 123 124 125 |
# File 'lib/perfect_shape/cubic_bezier_curve.rb', line 121 def point_crossings(x_or_point, y = nil, level = 0) x, y = normalize_point(x_or_point, y) return unless x && y CubicBezierCurve.point_crossings(points[0][0], points[0][1], points[1][0], points[1][1], points[2][0], points[2][1], points[3][0], points[3][1], x, y, level) end |
#point_segment_distance(x_or_point, y = nil, minimum_distance_threshold: OUTLINE_MINIMUM_DISTANCE_THRESHOLD) ⇒ Object
180 181 182 183 184 185 186 187 188 189 190 191 192 193 194 195 196 197 198 199 200 201 202 203 204 205 206 |
# File 'lib/perfect_shape/cubic_bezier_curve.rb', line 180 def point_segment_distance(x_or_point, y = nil, minimum_distance_threshold: OUTLINE_MINIMUM_DISTANCE_THRESHOLD) x, y = normalize_point(x_or_point, y) return unless x && y point = Point.new(x, y) current_curve = self minimum_distance = point.point_distance(curve_center_point) last_minimum_distance = minimum_distance + 1 # start bigger to ensure going through loop once at least while minimum_distance >= minimum_distance_threshold && minimum_distance < last_minimum_distance curve1, curve2 = current_curve.subdivisions distance1 = point.point_distance(curve1.curve_center_point) distance2 = point.point_distance(curve2.curve_center_point) last_minimum_distance = minimum_distance if distance1 < distance2 minimum_distance = distance1 current_curve = curve1 else minimum_distance = distance2 current_curve = curve2 end end if minimum_distance < minimum_distance_threshold minimum_distance else last_minimum_distance end end |
#subdivisions(number = 2) ⇒ Object
Subdivides CubicBezierCurve exactly at its curve center
returning 2 CubicBezierCurve's as a two-element Array by default
number parameter may be specified as an even number in case more
subdivisions are needed. If an odd number is given, it is rounded
up to the closest even number above it (e.g. 3 becomes 4).
147 148 149 150 151 152 153 154 155 156 157 158 159 160 161 162 163 164 165 166 167 168 169 170 171 172 173 174 175 176 177 178 |
# File 'lib/perfect_shape/cubic_bezier_curve.rb', line 147 def subdivisions(number = 2) number = (number.to_i / 2.0).ceil*2 x1 = points[0][0] y1 = points[0][1] ctrlx1 = points[1][0] ctrly1 = points[1][1] ctrlx2 = points[2][0] ctrly2 = points[2][1] x2 = points[3][0] y2 = points[3][1] centerx = (ctrlx1 + ctrlx2) / 2.0 centery = (ctrly1 + ctrly2) / 2.0 ctrlx1 = (x1 + ctrlx1) / 2.0 ctrly1 = (y1 + ctrly1) / 2.0 ctrlx2 = (x2 + ctrlx2) / 2.0 ctrly2 = (y2 + ctrly2) / 2.0 ctrlx12 = (ctrlx1 + centerx) / 2.0 ctrly12 = (ctrly1 + centery) / 2.0 ctrlx21 = (ctrlx2 + centerx) / 2.0 ctrly21 = (ctrly2 + centery) / 2.0 centerx = (ctrlx12 + ctrlx21) / 2.0 centery = (ctrly12 + ctrly21) / 2.0 default_subdivisions = [ CubicBezierCurve.new(points: [x1, y1, ctrlx1, ctrly1, ctrlx12, ctrly12, centerx, centery]), CubicBezierCurve.new(points: [centerx, centery, ctrlx21, ctrly21, ctrlx2, ctrly2, x2, y2]) ] if number > 2 default_subdivisions.map { |curve| curve.subdivisions(number - 2) }.flatten else default_subdivisions end end |